from typing import FrozenSet

WorkerSpecialization = str

DRIVER: WorkerSpecialization = "driver"
FITTER: WorkerSpecialization = "fitter"
HANDYMAN: WorkerSpecialization = "handyman"
ELECTRICIAN: WorkerSpecialization = "electrician"
MANAGER: WorkerSpecialization = "manager"
ENGINEER: WorkerSpecialization = "engineer"

WORKER_TYPES: FrozenSet[WorkerSpecialization] = frozenset({DRIVER, FITTER, HANDYMAN, ELECTRICIAN, MANAGER, ENGINEER})


import time
from typing import Union, List, Callable

from sampo.scheduler.base import SchedulerType, Scheduler
from sampo.scheduler.genetic.base import GeneticScheduler
from sampo.scheduler.heft.base import HEFTScheduler
from sampo.scheduler.heft_between.base import HEFTBetweenScheduler
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.topological.base import TopologicalScheduler
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.resource_cost import schedule_cost


def get_scheduler_ctor(scheduling_algorithm_type: SchedulerType) \
        -> Callable[[SchedulerType, ResourceOptimizer, WorkTimeEstimator], Scheduler]:
    if scheduling_algorithm_type is SchedulerType.HEFTAddBetween:
        return HEFTBetweenScheduler
    if scheduling_algorithm_type is SchedulerType.HEFTAddEnd:
        return HEFTScheduler
    if scheduling_algorithm_type is SchedulerType.Topological:
        return TopologicalScheduler
    return GeneticScheduler


def generate_schedule(scheduling_algorithm_type: SchedulerType,
                      work_time_estimator: WorkTimeEstimator,
                      work_graph: WorkGraph,
                      contractors: Union[Contractor, List[Contractor]],
                      validate_schedule: bool) -> Schedule:
    scheduler = get_scheduler_ctor(scheduling_algorithm_type)(work_estimator=work_time_estimator)
    start_time = time.time()
    schedule = scheduler.schedule(work_graph,
                                  [contractors] if isinstance(contractors, Contractor) else contractors,
                                  validate=validate_schedule)
    print(f'Time: {(time.time() - start_time) * 1000} ms')
    print(f'Cost: {schedule_cost(schedule)}')
    return schedule



from datetime import datetime, timedelta
from typing import Union, Optional

from sampo.schemas.time import Time


def ftime(dt: datetime, f: str = '%y-%m-%d %H:%M:%S') -> str:
    return dt.strftime(f)


def parse_datetime(dts: str, f: Optional[str] = None) -> datetime:
    """
    Parses datetime from string.
    :param dts: String datetime
    :param f: String format. If not provided, '%Y-%m-%d' and then '%y-%m-%d %H:%M:%S' are tried.
    :return:
    """
    if f is None:
        try:
            return datetime.strptime(dts, '%Y-%m-%d')
        except ValueError:
            return datetime.strptime(dts, '%y-%m-%d %H:%M:%S')
    return datetime.strptime(dts, f)


def add_time_delta(base_datetime: Union[datetime, str],
                   time_delta: Union[Time, float],
                   time_units: Optional[str] = 'days') -> datetime:
    """
    Adds time delta to base datetime.
    :param base_datetime:
    :param time_delta:
    :param time_units: can be days, seconds, microseconds, milliseconds, minutes, hours, weeks
    :return:
    """
    base = parse_datetime(base_datetime) if isinstance(base_datetime, str) else base_datetime
    delta = timedelta(**{time_units: int(time_delta.value if isinstance(time_delta, Time) else time_delta)})

    return base + delta
